program PascalSLex(input, output, srcfil);
(*        N. Wirth, E.T.H CH-8092 Zurich      *)
label 99;                (* escape when input consumed *)

const nkw   =   27;      (* no. of key words *)
      alng  =   10;      (* no. of significant chars in identifiers *)
      blankID = '          '; (* blank string of length alng *)
      llng  =   80;      (* input line length *)
      emax  =  308;      (* max exponent of real numbers *)
      emin  = -324;      (* min exponent *)
      kmax  =   15;      (* max no. of significant digits *)
      smax  =   78;      (* max size of string *)
      ermax =   58;      (* max error no. *)
      nmax  = maxint;    (* largest integer value on this machine *)
      CRLF  = TRUE;      (* true, if palatform uses CR/LF *)

type  (* The token types recognized by lex analyzer*)
      symbol = (intcon, realcon, charcon, stringt,
                notsy, plus, minus, times, idiv, rdiv, imod, andsy, orsy,
                egl, neg, gtr, geg, lss, leg,
                lparent, rparent, lbrack, rbrack, comma, semicolon, period,
                colon, becomes, constsy, typesy, varsy, functionsy,
                proceduresy, arraysy, recordsy, programsy, ident,
                beginsy, ifsy, casesy, repeatsy, whilesy, forsy,
                endsy, elsesy, untilsy, ofsy, dosy, tosy, downtosy, thensy);

       (* maximum string needed for identifier or keyword *)
       alfa = packed array [1..alng] of char;

var    sy: symbol;         (* last symbol read by insymbol *)
       id: alfa;           (* identifier from insymbol *)
       inum: integer;      (* integer from insymbol *)
       rnum: real;         (* real number from insymbol *)
       ch: char;           (* last character read from source program *)
       line: array [1..llng] of char;
       cc: integer;        (* character counter *)
       lc: integer;        (* program location counter *)
       ll: integer;        (* length of current line *)
       errs: set of 0..ermax; (* retains list of errors encountered *)
       errpos: integer;     (* error position for lexical error just found*)
       progname: array[1..20] of char; (* input file name *)
       key: array [1..nkw] of alfa; (* set of keywords *)
       ksy: array [1..nkw] of symbol; (* symbols associated with keywords *)
       sps: array [char] of symbol; (* special symbols *)

       st: packed array [0..smax] of char; (* string from insymbol *)

       srcfil: text; { source input file }

       synames: array[symbol] of alfa; (* strings names for symbols *)

procedure nextch; (* read next character; process line end *)
const TAB=9; charPerTab = 8;
begin
   if cc = ll then begin
      if eof(srcfil) then begin
         writeln;
         writeln(' Source completed');
         goto 99
      end;
      if errpos <> 0 then begin
         writeln; errpos := 0
      end;
      writeln;
      write(lc:5, '  '); lc := lc+1;
      ll := 0; cc := 0;
      while not eoln(srcfil) do begin
         read(srcfil, ch);
         if ch >= ' ' then begin
            ll := ll+1; write(ch);
            line[ll] := ch
         end
         else if ord(ch)=TAB then (* add blanks to next tab *)
            repeat ll := ll+1; write(' '); line[ll] := ' '
            until (ll mod charPerTab) = 1
      end;
      writeln; ll := ll+1; read(srcfil, line[ll]);
      if CRLF then read(srcfil, line[ll])  (* LF CR on PC *)
   end;
   cc := cc+1; ch := line[cc];
end (* nextch *);

procedure error(n: integer);  (* position carat (^) under error *)
begin
   if errpos = 0 then write(' ****');
   if cc > errpos then begin
      write(' ': cc-errpos, '^', n:2);
      errpos := cc+3; errs := errs + [n]
   end
end (* error *);

procedure insymbol;   (* reads next symbol *)
   label 1, 2, 3;
   var i, j, k, l, e: integer;

   procedure readscale;
      var s, sign: integer;
   begin nextch; sign := 1; s := 0;
      if ch = '+' then nextch
      else if ch = '-' then begin
         nextch; sign := -1
      end;
      while ch in ['0'..'9'] do begin
         s := 10*s + ord(ch) - ord('0'); nextch
      end;
      e := s*sign + e
   end (* readscale *);

   procedure adjustscale;
      var s: integer; d, t: real;
   begin
      if k+e > emax then error(21)
      else if k+e < emin then rnum := 0
      else begin
         s := abs(e); t := 1.0; d := 10.0;
         repeat
            while not odd(s) do begin
               s := s div 2; d := sqr(d)
            end;
            s := s-1; t := d*t
         until s = 0;
         if e >= 0 then rnum := rnum*t else rnum := rnum/t
      end
   end (* adjustscale *);

begin (* insymbol *)
1: while ch <= ' ' do nextch;
   if ch in ['A'..'Z', 'a'..'z'] then
   begin (* word *)
      k := 0; id := blankID;
      repeat
         if k < alng then begin
            k := k+1;
            if ch in ['A'..'Z'] then ch := chr(ord(ch)+32); (* lowercase *)
            id[k] := ch
         end;
         nextch
      until not (ch in ['A'..'Z', 'a'..'z', '0'..'9']);
      i := 1; j := nkw; (* binary search for keyword *)
      repeat l := (i+j) div 2;
         if id <= key[l] then j := l-1;
         if id >= key[l] then i := l+1
      until i > j;
      if i-1 > j then sy := ksy[l] (* found in keyword table *)
      else  sy := ident (* identifier since not a reserved word *)
   end else
   if ch in ['0'..'9'] then
   begin (* number *)
      k := 0; inum := 0; sy := intcon;  (* assume it's an integer number *)
      repeat inum := inum*10 + ord(ch) - ord('0');
         k := k+1; nextch
      until not (ch in ['0'..'9']);
      if (k > kmax) or (inum > nmax) then
         begin error(21); inum := 0; k := 0 end;
      if ch = '.' then (* real number *)
      begin nextch;
         if ch = '.' then ch := ':' else
            begin sy := realcon; rnum := inum; e := 0;
               while ch in ['0'..'9'] do begin
                  e := e-1;
                  rnum := 10.0*rnum + (ord(ch)-ord('0')); nextch
               end;
               if ch = 'e' then readscale;
               if e <> 0 then adjustscale
            end
      end
      else if ch = 'e' then begin (* real number *)
         sy := realcon; rnum := inum; e := 0;
         readscale; if e <> 0 then adjustscale;
      end
   end else
   case ch of
':': begin nextch;
         if ch = '=' then
           begin sy := becomes; nextch end  (* := *)
           else sy := colon
     end;
'<': begin nextch;
        if ch = '=' then begin sy := leg; nextch end
        else if ch = '>' then begin sy := neg; nextch end
        else sy := lss
     end;
'>': begin nextch;
         if ch = '=' then begin sy := geg; nextch end
         else sy := gtr
     end;
'.': begin nextch;
        if ch = '.' then
           begin sy := colon; nextch end
           else sy := period
     end;
'''': begin k := 0;
    2: nextch;
       if ch = '''' then begin
          nextch; if ch <> '''' then goto 3
       end;
       if k <= smax then begin
          k := k+1; st[k] := ch
       end;
       goto 2;
    3: if k = 1 then begin
          sy := charcon; inum := ord(st[1])
       end
       else if k = 0 then begin
          error(38); sy := charcon; inum := 0
       end
       else begin
          sy := stringt; inum := k;
       end
      end;
'(': begin nextch;
        if ch <> '*' then begin sy := lparent end
        else begin (* comment *)
           nextch;
           repeat
              while ch <> '*' do nextch;
              nextch
           until ch = ')';
           nextch; goto 1
        end
     end;
'+', '-',  '*', '/', ')', '=', ',', '[', ']', '#', '&', ';':
     begin
        sy := sps[ch]; nextch
     end;
'$', '%', '@', '\', '~', '{', '}', '^':
     begin
        error(24); nextch; goto 1
     end
  end; (* case ch *)
  write(synames[sy],' ')
end (* insymbol *);

procedure init;
begin
  key[ 1] := 'and       '; key[ 2] := 'array     ';
  key[ 3] := 'begin     '; key[ 4] := 'case      ';
  key[ 5] := 'const     '; key[ 6] := 'div       ';
  key[ 7] := 'do        '; key[ 8] := 'downto    ';
  key[ 9] := 'else      '; key[10] := 'end       ';
  key[11] := 'for       '; key[12] := 'function  ';
  key[13] := 'if        '; key[14] := 'mod       ';
  key[15] := 'not       '; key[16] := 'of        ';
  key[17] := 'or        '; key[18] := 'procedure ';
  key[19] := 'program   '; key[20] := 'record    ';
  key[21] := 'repeat    '; key[22] := 'then      ';
  key[23] := 'to        '; key[24] := 'type      ';
  key[25] := 'until     '; key[26] := 'var       ';
  key[27] := 'while     ';
  ksy[ 1] := andsy;        ksy[ 2] := arraysy;
  ksy[ 3] := beginsy;      ksy[ 4] := casesy;
  ksy[ 5] := constsy;      ksy[ 6] := idiv;
  ksy[ 7] := dosy;         ksy[ 8] := downtosy;
  ksy[ 9] := elsesy;       ksy[10] := endsy;
  ksy[11] := forsy;        ksy[12] := functionsy;
  ksy[13] := ifsy;         ksy[14] := imod;
  ksy[15] := notsy;        ksy[16] := ofsy;
  ksy[17] := orsy;         ksy[18] := proceduresy;
  ksy[19] := programsy;    ksy[20] := recordsy;
  ksy[21] := repeatsy;     ksy[22] := thensy;
  ksy[23] := tosy;         ksy[24] := typesy;
  ksy[25] := untilsy;      ksy[26] := varsy;
  ksy[27] := whilesy;
  sps['+'] := plus;        sps['-'] := minus;
  sps['*'] := times;       sps['/'] := rdiv;
  sps['('] := lparent;     sps[')'] := rparent;
  sps['='] := egl;         sps[','] := comma;
  sps['['] := lbrack;      sps[']'] := rbrack;
  sps['#'] := neg;         sps['&'] := andsy;
  sps[';'] := semicolon;

  synames[andsy] := 'andsy';         synames[arraysy] := 'arraysy';
  synames[beginsy] := 'beginsy';     synames[casesy] := 'casesy';
  synames[constsy] := 'constsy';     synames[idiv] := 'idiv';
  synames[dosy] := 'dosy';           synames[downtosy] := 'downtosy';
  synames[elsesy] := 'elsesy';       synames[endsy] := 'downtosy';
  synames[forsy] := 'forsy';         synames[functionsy] := 'functionsy';
  synames[ifsy] := 'ifsy';           synames[imod] := 'imod';
  synames[notsy] := 'notsy';         synames[ofsy] := 'ofsy';
  synames[orsy] := 'orsy';           synames[proceduresy] := 'procedursy';
  synames[programsy] := 'programsy'; synames[recordsy] := 'recordsy';
  synames[repeatsy] := 'repeatsy';   synames[thensy] := 'thensy';
  synames[tosy] := 'tosy';           synames[typesy] := 'typesy';
  synames[untilsy] := 'untilsy';     synames[varsy] := 'varsy';
  synames[whilesy] := 'whilesy';

  synames[intcon] := 'intcon';       synames[realcon] := 'realcon';
  synames[charcon] := 'charcon';     synames[stringt] := 'stringt';
  synames[plus] := 'plus';           synames[minus] := 'minus';
  synames[times] := 'times';         synames[rdiv] := 'rdiv';
  synames[egl] := 'egl';             synames[neg] := 'neg';
  synames[gtr] := 'gtr';             synames[geg] := 'geg';
  synames[lss] := 'lss';             synames[leg] := 'leg';
  synames[comma] := 'comma';         synames[period] := 'period';
  synames[semicolon] := 'semicolon'; synames[colon] := 'colon';
  synames[ident] := 'ident';         synames[becomes] := 'becomes';
  synames[andsy] := 'andsy';         synames[orsy] := 'orsy';
  synames[lparent] := 'lparent';     synames[rparent] := 'rparent';
  synames[lbrack] := 'lbrack';       synames[rbrack] := 'rbrack';

  lc := 0; ll := 0; cc := 0; ch := ' ';
  errpos := 0; errs := [];
end; (* init *)

begin (* main program *)
  writeln;
  writeln('Pascal-S compiler/interpreter');

  write('Enter name of file to be compiled: ');
  readln(progname);
  assign(srcfil,progname);
  reset(srcfil);

  init;
  while true do insymbol;
99:
  readln;
end. (* PascalSLex *)
